# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import pickle
from . import representation
from . import decorators
from . import misc
from . import cache
from . import registry
from . import colors
from . import plots
from . import stream_filters
from . import abstract_method

__all__ = (*pickle.__all__,
           *representation.__all__,
           *decorators.__all__,
           *misc.__all__,
           *cache.__all__,
           *registry.__all__,
           *colors.__all__,
           *plots.__all__,
           *stream_filters.__all__,
           *abstract_method.__all__,
)

from .pickle import *
from .representation import *
from .decorators import *
from .misc import *
from .cache import *
from .registry import *
from .colors import *
from .plots import *
from .stream_filters import *
from .abstract_method import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .utils import colors

__all__ = (
    'StampedKeyError',
    'UndefinedChemicalAlias',
    'UndefinedChemical',
    'UndefinedPhase',
    'DimensionError',
    'InfeasibleRegion',
    'InvalidMethod',
    'NoEquilibrium',
    'UnitInheritanceError',
    'message_with_object_stamp',
    'try_method_with_object_stamp',
    'raise_error_with_object_stamp'
)

class StampedKeyError(KeyError):
    def __init__(self, msg):
        super().__init__(msg)
    
    def __repr__(self):
        return f"KeyError({self.args[0]})"
    
    def __str__(self):
        return self.args[0]

StampedKeyError.__name__ = 'KeyError'

class InfeasibleRegion(RuntimeError):
    """Runtime error regarding infeasible processes."""
    def __init__(self, region, msg=None): 
        self.region = region
        if msg is None: msg = region + ' is infeasible'
        super().__init__(msg)

class UndefinedChemicalAlias(Exception):
    """Exception regarding undefined chemicals."""
    def __init__(self, ID): 
        self.ID = ID
        super().__init__(repr(ID))
    
UndefinedChemical = UndefinedChemicalAlias # Alias for backwards compatibility
    
class UndefinedPhase(Exception):
    """Exception regarding undefined phases."""
    def __init__(self, phase): super().__init__(repr(phase))

class NoEquilibrium(Exception):
    """Exception regarding an attempt to solve phase equilibrium when not applicable."""

class DimensionError(ValueError):
    """ValueError regarding wrong dimensions."""

class InvalidMethod(ValueError):
    """ValueError regarding an attempt to evaluate an invalid method."""
    def __init__(self, method):
        super().__init__(repr(method))
    
class UnitInheritanceError(NotImplementedError):
    """Exception regarding unit inhearitance."""
    
def message_with_object_stamp(object, msg):
    object_name = str(repr(object))
    if object_name in msg:
        return msg
    else:
        return colors.violet(object_name) + ' ' + msg

def raise_error_with_object_stamp(object, error):
    try: 
        msg, *args = error.args
        error.args = (message_with_object_stamp(object, msg), *args)
    except: pass
    raise error

def try_method_with_object_stamp(object, method, args=()):
    try:
        return method(*args)
    except StampedKeyError as error:
        raise StampedKeyError(message_with_object_stamp(object, error.args[0]))
    except KeyError as error:
        raise StampedKeyError(message_with_object_stamp(object, repr(error.args[0])))
    except Exception as error:
        raise_error_with_object_stamp(object, error)


# -*- coding: utf-8 -*-
"""
"""
from typing import NamedTuple
from warnings import warn
from graphviz import Digraph, Source
from IPython import display as _display
from collections import deque
from colorpalette import Color
from scipy.interpolate import interp1d
from matplotlib import pyplot as plt
from itertools import product
from PIL import Image
import contextlib
import pydot
import numpy as np
import imageio
import os
from matplotlib import colormaps
from matplotlib.colors import rgb2hex
import shutil

viridis = colormaps['viridis']

__all__ = (
    'PhenomeNode',
    'EquationNode',
    'VariableNode',   
    'PhenomeGraph',
    'BipartitePhenomeGraph',
    'PhenomenaGraph',
    'DotFile',
    'Edge',
    'all_subgraphs',
)

#: [dict] Default variable node settings
variable_node_options = {
    'gradientangle': '0',
    'width': '0.15',
    'height': '0.15',
    'orientation': '0.0',
    'peripheries': '1',
    'penwidth': '3',
    'margin': '0',
    'fillcolor': 'none',
    'shape': 'circle',
    'style': 'filled',
    # 'color': '#b4b1ae',
    'label': '',
    'pin': 'true',
}

#: [dict] Default equation node settings
equation_node_options = {
    'shape': 'circle',
    'style': 'filled',
    'gradientangle': '0',
    'width': '0.2',
    'height': '0.2',
    'orientation': '0.0',
    'peripheries': '1',
    'margin': '0',
    'label': '',
    'pin': 'true',
}

#: [dict] Default phenomenode settings
phenomenode_options = {
    'shape': 'box',
    'style': 'filled',
    'gradientangle': '0',
    'width': '0.2',
    'orientation': '0.0',
    'peripheries': '0',
    'margin': '0',
    'label': '',
    'pin': 'true',
}

#: [dict] Default phenomenode edge settings
phenomenode_edge_options = dict(
    label='', 
    arrowtail='none', 
    arrowhead='none', 
    minlen='1',
    dir='none',
    penwidth='3',
    style='solid',
    pin='true',
)

#: [dict] Default equation node colors by category
colors = {
    'material': '#a280b9', # purple
    'energy': '#ed586f', # red
    'vle': '#60c1cf', # blue
    'lle': '#79bf82', # green
    'reaction': '#dd7440', # orange
    'shortcut': '#00a996', # darker blue 
    'compression': '#5c5763', # black
}
edge_options = dict(
    label='', 
    arrowtail='none', 
    arrowhead='none', 
    headport='c', 
    tailport='c', 
    len='0.5',
    # minlen='0.5',
    dir='none',
    penwidth='3',
    style='solid',
    # pin='true',
)
folder = os.path.dirname(__file__)
folder = os.path.join(folder, 'temporary_images')
             
all_subgraphs = ('material_balance', 'energy_balance', 'phenomena')

def filter_nodes(nodes, cls):
    filtered_nodes = []
    past_nodes = set()
    for i in nodes:
        if i is None: continue
        if i in past_nodes: continue    
        if not isinstance(i, cls):
            raise ValueError('equation nodes can only be connected to variable nodes')
        filtered_nodes.append(i)
        past_nodes.add(i)
    return filtered_nodes

def group_equations(equations, criteria, all_equations, all_neighbors):
    for eq in equations:
        if eq in all_equations or eq in all_neighbors: continue
        if criteria.match(eq.name): 
            all_equations.add(eq)
            group_equations(eq.neighbors, criteria, all_equations, all_neighbors)
        else:
            all_neighbors.add(eq)

def get_order(values):
    N = len(values)
    order = [*range(N)]
    order = sorted(order, key=lambda x: values[x])
    past_values = {}
    for n, i in enumerate(order):
        value = values[i]
        if value in past_values:
            order[n] = past_values[value]
        else:
            past_values[value] = i
    return order

class EquationNode:
    __slots__ = ('name', 'inputs', 'outputs', 'tracked_outputs')
    
    def __init__(self, name):
        self.name = name    
    
    @property
    def category(self):
        name = self.name
        for category in colors:
            if category in name: return category
        raise AttributeError('no category')
    
    def set_equations(self, inputs=(), outputs=(), tracked_outputs=None):
        self.inputs = filter_nodes(inputs, VariableNode)
        self.outputs = filter_nodes(outputs, VariableNode)
        if tracked_outputs is None: 
            self.tracked_outputs = self.outputs
        else:
            self.tracked_outputs = filter_nodes(tracked_outputs, VariableNode)
        for i in self.variables: i.equations.append(self)
        
    @property
    def neighbors(self):
        return set([j for i in self.variables for j in i.equations])
        
    @property
    def variables(self):
        return (*self.inputs, *self.outputs)
    
    def get_edges(self, inputs=True, outputs=True):
        if inputs:
            nodes = self.inputs.copy()
            if outputs:
                nodes += self.outputs
        elif outputs:
            nodes = self.outputs
        else:
            nodes = ()
        return [Edge(self, i) for i in nodes]
        
    def __repr__(self):
        return self.name
    
    
class VariableNode:
    __slots__ = ('name', 'getter', 'equations', 'last_value')
    
    def __init__(self, name, getter):
        if not callable(getter):
            raise ValueError('getter must be function')
        self.name = name
        self.getter = getter
        self.equations = []
        self.last_value = np.nan
        
    def get_value(self):
        self.last_value = value = self.getter()
        return value
        
    def __repr__(self):
        return self.name


class Edge(NamedTuple):
    equation: EquationNode
    variable: VariableNode
    
    @property
    def name(self):
        return (self.equation.name, self.variable.name)
    

class Criteria:
    __slots__ = ('names',)
    
    def __init__(self, names):
        if isinstance(names, str): names = [names]
        self.names = names
        
    def match(self, full_name):
        return all([i in full_name for i in self.names])


class CriteriaSelection:
    __slots__ = ('criterias',)
    def __init__(self, selection):
        self.criterias = [Criteria(i) for i in selection]

    def get_criteria(self, full_name):
        for criteria in self.criterias:
            if criteria.match(full_name): return criteria

class PhenomeNode:
    __slots__ = ('name', 'size', 'equations', 'neighbors', 'criteria')
    
    def __init__(self, equation, criteria_selection):
        self.name = equation.name # Good enough
        self.equations = equations = set([equation])
        self.neighbors = neighbors = set()
        self.criteria = criteria = criteria_selection.get_criteria(equation.name)
        group_equations(equation.neighbors, criteria, equations, neighbors)
        
    category = EquationNode.category
        
    def finalize(self, stage_tags):
        name = '_'.join([i.name for i in self.equations])
        self.size = sum([i in name for i in stage_tags])
        
    def output_variables(self):
        return set([j for i in self.equations for j in i.tracked_outputs]) 
        
    def __repr__(self):
        return self.name


class PhenomeGraph:
    __slots__ = ('phenomenodes', 'edges', 'pydot', 'profiles', 'time')
    
    def __init__(self, equations, criteria_selection, stage_tags, variable_profiles, time, file=None):
        phenomenodes = []
        collected_equations = set()
        if not isinstance(criteria_selection, CriteriaSelection): criteria_selection = CriteriaSelection(criteria_selection)
        for eq in equations:
            if eq in collected_equations: continue
            phenomenode = PhenomeNode(eq, criteria_selection)
            phenomenodes.append(phenomenode)
            collected_equations.update(phenomenode.equations)
        reference = {
            eq: node for node in phenomenodes
            for eq in node.equations
        }
        edges = set([
            frozenset([node, reference[eq]]) for node in phenomenodes
            for eq in node.neighbors
        ])
        for i in phenomenodes: i.finalize(stage_tags)
        errors = np.abs(variable_profiles - variable_profiles.iloc[-1])
        def normalized_profile(errors):
            order = np.array(get_order(errors.sum(axis=1)))
            order_norm = 100 * order / order.max()
            # breakpoint()
            return order_norm
            # MSE = errors.sum(axis=1)
            # # MSE = (errors * errors).mean(axis=1)
            # mask = np.isnan(MSE)
            # MSE[mask] = MSE[~mask].max() # Maximum error
            # MSE[MSE < 1e-16] = 1e-16 # Minimum error
            # MSE = np.log(MSE)
            # MSE_zeroed = MSE - MSE.min()
            # MSE_norm = 100 * MSE_zeroed / MSE_zeroed.max()
            # return MSE_norm
        
        assert variable_profiles.shape[0] == len(time)
        profiles = [
            normalized_profile(errors[[i.name for i in phenomenode.output_variables()]].values)
            for phenomenode in phenomenodes
        ]
        self.time = time
        assert len(profiles[0]) == len(time)
        self.phenomenodes = phenomenodes
        self.edges = edges
        self.profiles = profiles
        self.load_pydot()
        
    def load_pydot(self, 
            file=None,
            maxiter=None,
            damping=None, 
            K=None,
        ):
        if file:
            try:
                self.pydot = pydot.graph_from_dot_file(file)[0]
            except:
                self.load_pydot(None, maxiter, damping, K)
                self.write(file)
            return
        if maxiter is None: maxiter = '20000'
        if damping is None: damping = '0.3'
        if K is None: K = '0.2'
        # Create a digraph and set direction left to right
        digraph = Digraph(format='png', strict=True)
        digraph.attr(
            'graph',
            rankdir='LR', 
            maxiter=maxiter, 
            Damping=damping, 
            K=K,
            penwidth='0', 
            color='none', 
            bgcolor='transparent',
            nodesep='0.02', 
            ranksep='0.02', 
            layout='dot', 
            splines='curved', 
            outputorder='edgesfirst', 
            dpi='300',
            # nodesep='0.2', 
            # ranksep='0.2',
            overlap='compress', 
            dir='none',
        )
        dct = {}
        for i in self.phenomenodes:
            names = i.criteria.names
            if len(names) == 1:
                key, = names
                if key in dct: dct[key].append(i)
                else: dct[key] = [i]
            else:
                outer, inner = names
                if outer in dct: subdct = dct[outer]
                else: dct[outer] = subdct = {}
                if inner in subdct: subdct[inner].append(i)
                else: subdct[inner] = [i]
            
        for m, subdct in enumerate(dct.values()):
            if isinstance(subdct, dict):
                with digraph.subgraph(name=f'cluster_{m}') as subgraph:
                    subgraph.attr('graph', color='none', bgcolor='none', shape='box', label='')
                    for n, phenomenodes in enumerate(subdct.values()): 
                        with subgraph.subgraph(name=f'cluster_{m}_{n}') as sub2:
                            sub2.attr('graph', color='none', bgcolor='none', shape='box', label='')
                            for i in phenomenodes:
                                color = colors[i.category]
                                sub2.node(
                                    name=i.name, 
                                    color='black',
                                    fillcolor=color,
                                    height=str(i.size * 0.2),
                                    **phenomenode_options
                                )
            else:
                n = m
                phenomenodes = subdct
                with digraph.subgraph(name='cluster_' + str(n)) as subgraph:
                    subgraph.attr('graph', color='none', bgcolor='none', shape='box', label='')
                    for i in phenomenodes:
                        color = colors[i.category]
                        subgraph.node(
                            name=i.name, 
                            color='black',
                            fillcolor=color,
                            height=str(i.size * 0.2),
                            **phenomenode_options
                        )
                                
        # Set attributes for graph and edges
        digraph.attr(
            'edge', 
            concentrate='true',
            dir='none'
        )
        for left, right in self.edges:
            if right.size == 1:
                color = colors[right.category]
            elif left.size == 1:
                color = colors[right.category]
            else:
                color = 'black'
            digraph.attr('edge', label='', taillabel='', headlabel='')
            digraph.edge(
                left.name, right.name, 
                color='black',
                **phenomenode_edge_options,
            )
        dot = digraph.pipe(format='dot') # Pin nodes
        self.pydot = pydot.graph_from_dot_data(dot.decode('utf-8'))[0]
        
    def display(self): # TODO: this does not work, maybe revert back to graphviz  
        source = Source(self.pydot.create_dot().decode('utf-8'))
        # img = source.pipe('svg')
        # x = _display.SVG(img)
        x = _display.Image(source.pipe(format='png'))
        return _display.display(x)
        
    def write(self, file):
        format = file[file.index('.')+1:]
        method = 'write_' + format
        getattr(self.pydot, method)(file)
        
    # TODO: plots/gifs
    # def plot_convergence_profile(self, file=None):
    #     equation_profiles = self.equation_profiles
    #     profiles_arr = np.abs(equation_profiles.values)
    #     equations = sorted(self.equations, key=lambda x: x.category)
    #     index = [equation_profiles.columns.get_loc(i.name) for i in equations]
    #     # ps = [*range(0, 101, 1)]
    #     # percentiles = np.percentile(profiles_arr, ps, axis=0)
    #     # fs = [interp1d(i, ps) for i in percentiles.T]
    #     M, N = profiles_arr.shape
    #     image = np.zeros([M, N, 3])
    #     # errors_percentiles = np.zeros([M, N])
    #     # for i in range(M):
    #     #     for j in index: 
    #     #         errors_percentiles[i, j] = fs[j](profiles_arr[i, j])
    #     # errors_percentiles -= errors_percentiles.min(axis=0, keepdims=True)
    #     profiles_arr[:] = np.log(profiles_arr + 1e-6)
    #     profiles_arr -= profiles_arr.min(axis=0, keepdims=True)
    #     max_errors = profiles_arr.max(axis=0, keepdims=True)
    #     max_errors[max_errors < 1e-9] = 1
    #     profiles_arr /= max_errors
    #     for i in range(M):
    #         for k, (j, eq) in enumerate(zip(index, equations)): 
    #             # error = errors_percentiles[i, j]
    #             # if error < 20: error = 0.
    #             error = 100 * profiles_arr[i, j]
    #             color = colors[eq.category]
    #             image[i, k, :] = Color(fg=color).shade(error).RGBn        
    #     ax = plt.imshow(image)
    #     plt.axis('off')
    #     if file:
    #         for i in ('svg', 'png'):
    #             plt.savefig(file, dpi=900, transparent=True)
    #     return ax
        
    def convergence_gif(
            self, 
            file=None, 
            total_duration=30, # Seconds
            fps=3,
            **kwargs
        ):
        digraph = self.pydot
        digraph.set_maxiter('10') # This speeds up image creation.
        digraph.set_dpi('100') # Pydot does not seem to save dpi from piped graphviz
        # breakpoint()
        output_file = file
        input_file = os.path.join(folder, 'temp0.png')
        profiles = self.profiles
        total_frames = total_duration * fps
        time = self.time
        normalized_time = time / time[-1] * total_duration
        frame_duration = 1 / fps 
        interpolators = [interp1d(normalized_time, profiles[name]) for name in profiles]
        phenomenodes = self.phenomenodes
        nodes_dct = {}
        for subgraph in digraph.get_subgraph_list():
            if self.depth == 1:
                for i in subgraph.get_nodes():
                    nodes_dct[i.get_name().strip('"')] = i
            elif self.depth == 2:
                for subgraph in subgraph.get_subgraph_list():
                    for i in subgraph.get_nodes():
                        nodes_dct[i.get_name().strip('"')] = i
        print('total frames', total_frames)
        t = 0
        for n in range(total_frames):
            print('Frame #', n)
            t += frame_duration
            for j, node in enumerate(phenomenodes): 
                error = 0.8 * interpolators[j](t) # Leave at least 20% of color in
                color = Color(fg=colors[node.category]).shade(error).HEX
                node = nodes_dct[node.name]
                node.set_fillcolor(color)
            digraph.write_png(input_file.replace('0', str(n)))
            
        if output_file is None: output_file = os.path.join(folder, 'temp.gif')
        elif '.' not in output_file: output_file += '.gif'
        
        with contextlib.ExitStack() as stack:
            # lazily load images
            imgs = (
                stack.enter_context(
                    Image.open(
                        input_file.replace('0', str(i))
                    )
                ) for i in range(total_frames)
            )
        
            # extract  first image from iterator
            img = next(imgs)
        
            # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
            img.save(fp=output_file, format='GIF', append_images=imgs,
                     save_all=True, duration=frame_duration * 1000, optimize=False, loop=1)
            

class BipartitePhenomeGraph:
    __slots__ = ('name', 'equations', 'variables', 'edges', 
                 'variable_profiles', 'equation_profiles', 'edge_profiles',
                 'subgraphs', 'pydot', 'time',)
    
    @staticmethod
    def get_node_edges(equation_nodes, inputs=True, outputs=True):
        return set(sum([i.get_edges(inputs, outputs) for i in equation_nodes], []))
    
    def __init__(self, 
            name, equations, variables, edges, 
            equation_profiles, variable_profiles, 
            edge_profiles, time, file=None,
            **kwargs,
        ):
        self.name = name
        self.equations = equations
        self.variables = variables
        self.edges = edges
        self.equation_profiles = equation_profiles
        self.variable_profiles = variable_profiles
        self.edge_profiles = edge_profiles
        self.time = time
        self.subgraphs = []
        self.load_pydot(file, **kwargs)
       
    def subgraph(self, 
            name, equations, variables, edges, 
            equation_profiles, variable_profiles, 
            edge_profiles, file=None, **kwargs,
        ):
        subgraph = BipartitePhenomeGraph.__new__(BipartitePhenomeGraph)
        subgraph.name = name
        subgraph.equations = equations
        subgraph.variables = variables
        subgraph.edges = edges
        subgraph.equation_profiles = equation_profiles
        subgraph.variable_profiles = variable_profiles
        subgraph.edge_profiles = edge_profiles
        subgraph.time = self.time
        self.subgraphs.append(subgraph)
        subgraph.load_pydot(file, parent=self.pydot, **kwargs)
       
    def load_pydot(self, 
            file=None,
            maxiter=None,
            damping=None, 
            K=None,
            subgraph_units=None,
            parent=None,
        ):
        if file:
            try:
                self.pydot = pydot.graph_from_dot_file(file)[0]
            except:
                self.load_pydot(None, maxiter, damping, K, subgraph_units, parent)
                self.write(file)
            return
        if maxiter is None: maxiter = '10000'
        if damping is None: damping = '0.2'
        if K is None: K = '0.2'
        # Create a digraph and set direction left to right
        digraph = Digraph(format='png', strict=True)
        digraph.attr(
            'graph',
            rankdir='LR', 
            maxiter=maxiter, 
            Damping=damping, 
            K=K,
            sep='1',
            penwidth='0', 
            color='none', 
            bgcolor='transparent',
            nodesep='0.02', 
            ranksep='0.02', 
            layout='fdp', 
            splines='curved', 
            outputorder='edgesfirst', 
            dpi='300',
            # nodesep='0.2', 
            # ranksep='0.2',
            overlap='compress', 
            dir='none',
        )
        variables = self.variables
        equations = self.equations
        edges = self.edges
        inner_edges = []
        past_nodes = set()
        subgraph_units = False
        if subgraph_units:
            for name in subgraph_units:
                # with digraph.subgraph(name=name) as subgraph:
                # subgraph.attr('graph', color='#b4b1ae', bgcolor='#b4b1ae', shape='box', layout='fdp')
                subgraph_equations = [i for i in equations if name in i.name]
                subgraph_variables = set(sum([i.outputs for i in subgraph_equations], []))
                subgraph_variables = set([i for i in subgraph_variables if name in i.name])
                assert subgraph_variables.isdisjoint(past_nodes)
                past_nodes.update(subgraph_variables)
                subgraph_edges = BipartitePhenomeGraph.get_node_edges(subgraph_equations)
                subgraph_edges = tuple([i for i in subgraph_edges if i.variable in subgraph_variables])
                subgraph_variables = tuple(subgraph_variables)
                inner_edges.extend(subgraph_edges)
                self.fill_digraph_nodes(digraph, subgraph_variables, subgraph_equations)
                self.fill_digraph_edges(digraph, subgraph_edges)
            self.fill_digraph_edges(digraph, set(edges).difference(inner_edges))
        else:
            self.fill_digraph_nodes(digraph, variables, equations)
            self.fill_digraph_edges(digraph, edges)
        # img = digraph.pipe(format='png') # Pin nodes
        # f = open(self.name + '.png', 'wb')
        # f.write(img)
        # f.close()
        dot = digraph.pipe(format='dot') # Pin nodes
        self.pydot = pydot.graph_from_dot_data(dot.decode('utf-8'))[0]
        # for node in self.pydot.get_nodes(): node.set_pin('false')
        if parent:
            nodes_dct = {i.get_name().strip('"'): i for i in self.pydot.get_nodes()}
            parent_nodes_dct = {i.get_name().strip('"'): i for i in parent.get_nodes()}
            edges_dct = {
                (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
                for i in self.pydot.get_edges()
            }
            parent_edges_dct = {
                (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
                for i in parent.get_edges()
            }
            for name, node in nodes_dct.items():
                pos = parent_nodes_dct[name].get_pos()
                if pos: node.set_pos(pos)
            for key, edge in edges_dct.items():
                pos = parent_edges_dct[key].get_pos()
                if pos: edge.set_pos(pos)
        
    def fill_digraph_nodes(self, digraph, variables, equations):
        variable_colors = {}
        for i in equations: 
            color = colors[i.category]
            digraph.node(
                name=i.name, 
                color=color,
                fillcolor=color,
                **equation_node_options
            )
            for var in i.outputs: variable_colors[var.name] = color
        for i in variables: 
            digraph.node(
                name=i.name, 
                color=variable_colors[i.name],
                **variable_node_options
            )
    
    def fill_digraph_edges(self, digraph, edges):
        # Set attributes for graph and edges
        digraph.attr(
            'edge', 
            dir='none'
        )
        equations = set(self.equations)
        variables = set(self.variables)
        for edge in edges:
            equation = edge.equation
            variable = edge.variable
            assert equation in equations
            assert variable in variables
            digraph.attr('edge', label='', taillabel='', headlabel='')
            color = colors[equation.category]
            digraph.edge(
                equation.name, variable.name, 
                color=color, name='_'.join(edge.name),
                **edge_options,
            )
        
    def display(self): # TODO: this does not work, maybe revert back to graphviz  
        source = Source(self.pydot.create_dot().decode('utf-8'))
        # img = source.pipe('svg')
        # x = _display.SVG(img)
        x = _display.Image(source.pipe(format='png'))
        return _display.display(x)
        
    def write(self, file):
        format = file[file.index('.')+1:]
        method = 'write_' + format
        getattr(self.pydot, method)(file)
        
    # def plot_convergence_profile(self, file=None):
    #     equation_profiles = self.equation_profiles
    #     profiles_arr = np.abs(equation_profiles.values)
    #     equations = sorted(self.equations, key=lambda x: x.category)
    #     index = [equation_profiles.columns.get_loc(i.name) for i in equations]
    #     # ps = [*range(0, 101, 1)]
    #     # percentiles = np.percentile(profiles_arr, ps, axis=0)
    #     # fs = [interp1d(i, ps) for i in percentiles.T]
    #     M, N = profiles_arr.shape
    #     image = np.zeros([M, N, 3])
    #     # errors_percentiles = np.zeros([M, N])
    #     # for i in range(M):
    #     #     for j in index: 
    #     #         errors_percentiles[i, j] = fs[j](profiles_arr[i, j])
    #     # errors_percentiles -= errors_percentiles.min(axis=0, keepdims=True)
    #     profiles_arr[:] = np.log(profiles_arr + 1e-6)
    #     profiles_arr -= profiles_arr.min(axis=0, keepdims=True)
    #     max_errors = profiles_arr.max(axis=0, keepdims=True)
    #     max_errors[max_errors < 1e-9] = 1
    #     profiles_arr /= max_errors
    #     for i in range(M):
    #         for k, (j, eq) in enumerate(zip(index, equations)): 
    #             # error = errors_percentiles[i, j]
    #             # if error < 20: error = 0.
    #             error = 100 * profiles_arr[i, j]
    #             color = colors[eq.category]
    #             image[i, k, :] = Color(fg=color).shade(error).RGBn        
    #     ax = plt.imshow(image)
    #     plt.axis('off')
    #     if file:
    #         for i in ('svg', 'png'):
    #             plt.savefig(file, dpi=900, transparent=True)
    #     return ax
        
    def convergence_gif(self, 
            reference, time, profiles, 
            file=None, total_duration=None, fps=None, 
            categorize=True, interpolate=False,
            inverse=False, **kwargs
        ):
        digraph = self.pydot
        equations = self.equations
        digraph.set_maxiter('10') # This speeds up image creation.
        digraph.set_dpi('100') # Pydot does not seem to save dpi from piped graphviz
        # breakpoint()
        output_file = file
        input_file = os.path.join(folder, 'temp0.png')
        time = self.time
        # profiles = self.variable_profiles
        # reference = {i: n for n, i in enumerate(profiles)}
        # profiles = profiles.values
        # profiles = np.abs(profiles - profiles[-1])
        # profiles -= profiles.min(axis=0, keepdims=True)
        # profiles *= 100 / profiles.max()
        pulses = (False, True, False)
        if interpolate:
            raise NotImplementedError('not implemented in BioSTEAM yet')
            if fps is None: fps = 3
            total_frames = total_duration * fps
            time -= time[0]
            normalized_time = time / time[-1] * total_duration
            frame_duration = 1 / fps 
            interpolators = [interp1d(normalized_time, i) for i in profiles]
            duration_ms = 1000 * frame_duration
            # print(duration_ms)
        else:
            # time = time[:total_frames + 1] - time[0]
            # normalized_time = time / time[-1] * total_duration
            # duration_ms = 1000 * np.diff(normalized_time) # in ms
            # duration_ms = [*duration_ms]
            # duration_ms = 1 / fps * 1000
            # print(duration_ms)
            # breakpoint()
            total_frames = len(profiles[0])
            if total_duration is not None:
                duration_ms = 1000 * total_duration / total_frames
                fps = duration_ms / 1000
            elif fps is not None:
                total_duration = total_frames / fps
            else:
                total_duration = 16
                fps = total_duration / total_frames
            duration_ms = 1000 / (fps * len(pulses))
            print('total_duration', total_duration)
            # profiles = [i[:total_frames] for i in profiles]
            # profiles = [i - i.min() for i in profiles]
            # profiles = [100 * i / (i.max() or 1) for i in profiles]
        total_frames = len(profiles[0])
        print('total frames', total_frames)
        variables = set(self.variables)
        nodes_dct = {i.get_name().strip('"'): i for i in digraph.get_nodes()}
        equation_names = set([i.name for i in equations])
        variable_names = set([i.name for i in variables])
        edges_dct = {
            (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
            for i in digraph.get_edges()
        }
        if interpolate: 
            t = 0
            dt = total_duration / (total_frames - 1)
        filenames = []
        done = set()
        for n in range(total_frames):
            print('Frame #', n)
            if interpolate: 
                t += dt
                if t > total_duration: t = total_duration # floating point error
            nfile = input_file.replace('0', str(n))
            for pulse in pulses: 
                pulse_nfile = nfile.replace('.png', '_pulse.png')
                if (n, pulse) in done:
                    if pulse: 
                        filenames.append(pulse_nfile)
                    else:
                        filenames.append(nfile)
                    continue
                for eq in equations:
                    index = reference[eq.name]
                    if interpolate: 
                        baseline = interpolators[index](t) 
                        if pulse and n:
                            last = interpolators[index](t - dt) 
                            error = max(1.5 * baseline - 0.5 * last, 0) # Leave at least 20% of color in
                        else:
                            error = baseline
                    else:
                        baseline = profiles[index][n]
                        if pulse and n:
                            last = profiles[index][n - 1]
                            error = max(1.5 * baseline - 0.5 * last, 0) # Leave at least 20% of color in
                        else:
                            error = baseline
                    error = 0.8 * error # Leave at least 20% of color in
                    color = colors[eq.category]
                    color = Color(fg=color).shade(error).HEX
                    node = nodes_dct[eq.name]
                    node.set_color(color)
                    node.set_fillcolor(color)
                    edges = [i for i in eq.get_edges() if i.variable in variables]
                    for edge in edges:
                        _edge = edges_dct[edge.equation.name, edge.variable.name]
                        _edge.set_color(color)
                    for var in eq.outputs:
                        var = nodes_dct[var.name]
                        var.set_color(color)
                if pulse: 
                    digraph.write_png(pulse_nfile)
                    filenames.append(pulse_nfile)
                else:
                    digraph.write_png(nfile)
                    filenames.append(nfile)
                done.add((n, pulse))
                
        if output_file is None: output_file = os.path.join(folder, 'temp.gif')
        elif '.' not in output_file: output_file += '.gif'
        with contextlib.ExitStack() as stack:
            # lazily load images
            imgs = (
                stack.enter_context(
                    Image.open(i)
                ) for i in filenames
            )
        
            # extract  first image from iterator
            img = next(imgs)
        
            duration = [duration_ms] * len(filenames)
            duration[-1] += 5000
            print(sum(duration) / 1000)
            # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
            img.save(fp=output_file, format='GIF', append_images=imgs, loop=0,
                     save_all=True, duration=duration, optimize=False)
            
    def __repr__(self):
        return self.name
    
PhenomenaGraph = BipartitePhenomeGraph # For backwards compatibility

class DotFile:
    __slots__ = ('file', 'subfiles')
    
    def __init__(self, file):
        self.file = file if file.endswith('.dot') else f'{file}.dot'
         
    def subfile(self, subgraph):
        return self.file.replace('.dot', f'_{subgraph}.dot')
        
    def __repr__(self):
        return f"{type(self).__name__}(file={self.file!r})"


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

__all__ = ('read_only',)

def deny(self, *args, **kwargs):
    raise TypeError(f"'{type(self).__name__}' object is read-only")

def read_only(cls=None, methods=()):
    if not cls and methods:
        return lambda cls: read_only(cls, methods)
    else:
        for i in methods: setattr(cls, i, deny)
        cls.__delattr__ = deny
        cls.__setattr__ = deny
        return cls
 


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the tp_dependent_property module from the thermo library:
# https://github.com/CalebBell/thermo
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from thermo.volume import COOLPROP, EOS, IDEAL, NEGLECT_P
from thermo import (
    TPDependentProperty,
    VaporPressure, 
    EnthalpyVaporization,
    SurfaceTension,
    SublimationPressure, EnthalpySublimation,
    VolumeSolid, VolumeLiquid, VolumeGas, VolumeSupercriticalLiquid,
    HeatCapacitySolid, HeatCapacityLiquid, HeatCapacityGas,
    ThermalConductivityLiquid, ThermalConductivityGas,
    ViscosityLiquid, ViscosityGas,
    PermittivityLiquid,
)

subs = (
    VaporPressure, EnthalpyVaporization,SurfaceTension,
    SublimationPressure, EnthalpySublimation,
    VolumeSolid, VolumeLiquid, VolumeGas, VolumeSupercriticalLiquid,
    HeatCapacitySolid, HeatCapacityLiquid, HeatCapacityGas,
    ThermalConductivityLiquid, ThermalConductivityGas,
    ViscosityLiquid, ViscosityGas, PermittivityLiquid,
)

# Remove cache from call
def __call__(self, T, P):
    if self.hook: 
        return self.hook(T, P)
    if self._method_P:
        return self.TP_dependent_property(T, P)
    else:
        return self.T_dependent_property(T)

TPDependentProperty.__call__ = __call__

# Missing method 
def has_method(self):
    return bool(self._method or self._method_P and self._method_P != NEGLECT_P)

TPDependentProperty.hook = None
TPDependentProperty.__bool__ = has_method

# Handling methods

@TPDependentProperty.method_P.setter
def method_P(self, method):
    if method is not None:
        method, *_ = method.split('(')
        method = method.upper().replace(' ', '_').replace('_AND_', '_').strip('_').replace('SOLID', 'S')
        if method not in self.all_methods_P and method != 'POLY_FIT':
            raise ValueError("Pressure dependent method '%s' is not available for this chemical; "
                             "available methods are %s" %(method, self.all_methods_P))
    self._method_P = method

TPDependentProperty.method_P = method_P

for cls in subs:
    try:
        methods = cls.ranked_methods
        methods.remove(COOLPROP)
        methods.append(COOLPROP)
    except: pass
    try: cls.ranked_methods_P.remove(COOLPROP)
    except: pass

VolumeLiquid.property_max = 1e6
VolumeSolid.property_max = 1e6
VolumeGas.ranked_methods_P.remove(IDEAL)
VolumeGas.ranked_methods_P.append(IDEAL)

for methods in (VolumeLiquid.ranked_methods, VolumeLiquid.ranked_methods_P, VolumeGas.ranked_methods_P):
    methods.remove(EOS)
    methods.append(EOS)

VolumeGas.RAISE_PROPERTY_CALCULATION_ERROR = False
def calculate_molar_volume_with_ideal_fallback(self, T, P):
    return self.TP_dependent_property(T, P) or self.calculate_P(T, P, IDEAL)

VolumeGas.__call__ = calculate_molar_volume_with_ideal_fallback


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .utils.decorators import registered_franchise
from ._stream import Stream

__all__ = ('Power',)

@registered_franchise(Stream)
class Heat:
    __slots__ = ('_ID', '_source', '_sink', 'heat')
    
    @property
    def source(self):
        return self._source
    
    @property
    def sink(self):
        return self._sink
    
    def isempty(self):
        return self.heat == 0.
    
    def empty(self):
        self.heat = 0.
    
    @classmethod
    def sum(cls, streams, ID=None):
        new = Heat(ID)
        new.mix_from(streams)
        return new
    
    def mix_from(self, streams):
        isa = isinstance
        heat = 0.
        HeatOrPower = (Power, Heat)
        for i in streams:
            if isa(i, HeatOrPower):
                heat += i.heat
            else:
                raise TypeError(f"can only mix from Heat or Power objects, not '{type(i).__name__}'")
        self.heat = heat
    
    def __init__(self, ID=None, heat=0., source=None, sink=None):
        self._register(ID)
        self.heat = heat #: heat [kJ / hr]
        self._source = source
        self._sink = sink
        
    def __repr__(self):
        if self._source and self._sink:
            return f"{type(self).__name__}(heat={self.heat:.2f}, source={self.source}, sink={self.sink})"
        elif self._source:
            return f"{type(self).__name__}(heat={self.heat:.2f}, source={self.source})"
        elif self._sink:
            return f"{type(self).__name__}(heat={self.heat:.2f}, sink={self.sink})"
        else:
            return f"{type(self).__name__}(heat={self.heat:.2f})"


@registered_franchise(Stream)
class Power:
    __slots__ = ('_ID', '_source', '_sink', 'power')
    
    @property
    def heat(self):
        return self.power # Power can be directly converted to heat without any loses.
    
    @property
    def source(self):
        return self._source
    
    @property
    def sink(self):
        return self._sink
    
    def isempty(self):
        return self.power == 0.
    
    def empty(self):
        self.power = 0.
    
    @classmethod
    def sum(cls, streams, ID=None):
        new = Power(ID)
        new.mix_from(streams)
        return new
    
    def mix_from(self, streams):
        isa = isinstance
        power = 0.
        for i in streams:
            if isa(i, Power):
                power += i.power
            else:
                raise TypeError(f"can only mix from Power objects, not '{type(i).__name__}'")
        self.power = power
    
    def __init__(self, ID=None, power=0., source=None, sink=None):
        self._register(ID)
        self.power = power #: Electric power [kJ / hr]
        self._source = source
        self._sink = sink
        
    def __repr__(self):
        if self._source and self._sink:
            return f"{type(self).__name__}(power={self.power:.2f}, source={self.source}, sink={self.sink})"
        elif self._source:
            return f"{type(self).__name__}(power={self.power:.2f}, source={self.source})"
        elif self._sink:
            return f"{type(self).__name__}(power={self.power:.2f}, sink={self.sink})"
        else:
            return f"{type(self).__name__}(power={self.power:.2f})"
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2025, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import thermosteam as tmo
from scipy.optimize import minimize
from typing import NamedTuple
from itertools import combinations

__all__ = ('TangentPlaneStabilityAnalysis',)

class StabilityReport(NamedTuple):
    unstable: bool
    candidate: np.ndarray
    tpd: float

def edge_points_simplex_masked(z: np.ndarray,
                               points_per_edge: int = 10,
                               epsilon: float = 1e-3,
                               min_active: int = 2) -> np.ndarray:
    """
    Sample along edges of the simplex, but only for components with z_i > mean(z).
    Inactive components keep their original composition (scaled so total = 1).
    Always uses Chebyshev spacing along edges (no pure vertices).

    Parameters
    ----------
    z : array_like
        Overall composition (mole fractions, will be normalized).
    points_per_edge : int
        Number of points along each edge.
    epsilon : float
        Small offset to avoid exact zeros (safe for logs/fugacities).
    min_active : int
        Minimum number of components to activate if few exceed average.

    Returns
    -------
    np.ndarray
        Array of shape (n_points, n_components).
    """
    z = np.asarray(z, float)
    z = z / z.sum()
    n = z.size

    # filter: components strictly greater than average
    avg = z.mean()
    active = np.where(z >= avg)[0]

    # fallback: ensure at least `min_active` components
    if active.size < min_active:
        order = np.argsort(-z)
        active = order[:min(n, min_active)]

    # Chebyshev nodes mapped to [0, 1]
    k = np.arange(points_per_edge)
    t = 0.5 * (1 - np.cos(np.pi * k / (points_per_edge - 1)))

    pts = []

    # edges only among the selected components
    for i, j in combinations(active, 2):
        for tau in t:
            w = np.zeros(n)

            # inactive components fixed to original z_i
            for idx in range(n):
                if idx not in (i, j):
                    w[idx] = z[idx]

            # fraction available for the edge = leftover after inactive
            inactive_sum = w.sum()
            free = 1.0 - inactive_sum

            # assign varying parts to i, j
            w[i] = (1 - tau) * free
            w[j] = tau * free

            # avoid exact zeros
            w = np.maximum(w, epsilon)
            w /= w.sum()

            pts.append(w)

    return np.array(pts)


class TangentPlaneStabilityAnalysis:
    
    def __init__(self, phases, chemicals, thermo=None):
        thermo = tmo.settings.get_default_thermo(thermo)
        self.fugacity_models =  {
            i: thermo.fugacities(i, chemicals)
            for i in phases
        }
    
    def objective(self, w, T, P, model, logfz, reduce, softmax=False):
        if softmax:
            w = np.exp(w - np.max(w)) # Softmax for unconstrained optimization
            w /= w.sum()
        return np.dot(w, np.log(model(w, T, P, reduce) + 1e-30) - logfz)
    
    def __call__(self, z, T, P, reference_phase='l', potential_phase='L'):
        reference_model = self.fugacity_models[reference_phase]
        same_phase = reference_phase.lower() == potential_phase.lower()
        logfz = np.log(reference_model(z, T, P, same_phase) + 1e-30)
        best_val = np.inf
        best_result = None
        samples = edge_points_simplex_masked(z)
        objective = self.objective
        model = self.fugacity_models[potential_phase]
        args = (T, P, model, logfz, same_phase)
        for sample in samples:
            value = objective(sample, *args)
            if value < best_val:
                best_val = value
                best_result = sample
        result = minimize(
            objective, 
            best_result, 
            method="L-BFGS-B", 
            options=dict(maxiter=5),
            args=(*args, True),
        )
        value = result.fun
        if value < best_val:
            w = result.x
            w = np.exp(w - np.max(w)) # Softmax for unconstrained optimization
            w /= w.sum()
            best_result = w
        return StabilityReport(
            unstable=(best_val < 0),
            candidate=best_result,
            tpd=best_val
        )


if __name__ == '__main__':
    tmo.settings.set_thermo(['Water', 'Octanol'])
    TPSA = TangentPlaneStabilityAnalysis('lL', tmo.settings.chemicals)
    report = TPSA(np.array([0.2, 0.8]), 298.15, 101325)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
import thermosteam as tmo
import numpy as np
from math import exp
from thermosteam import functional as fn
from thermo.interaction_parameters import IPDB
from numpy.typing import NDArray
from thermo import eos_mix
from .. import units_of_measure as thermo_units
from ..base import PhaseHandle, MockPhaseTHandle, MockPhaseTPHandle, SparseVector, sparse
from .ideal_mixture_model import (
    SinglePhaseIdealTMixtureModel,
    IdealTMixtureModel, 
    IdealTPMixtureModel, 
    IdealEntropyModel, 
    IdealHvapModel
)
from .._chemicals import Chemical, CompiledChemicals, chemical_data_array

__all__ = ('Mixture', 'IdealMixture')

# %% Functions for building mixture models

def create_mixture_model(chemicals, var, Model):
    getfield = getattr
    isa = isinstance
    handles = []
    for chemical in chemicals:
        obj = getfield(chemical, var)
        if isa(obj, PhaseHandle):
            phase_handle = obj
        elif var == 'Cn':
            phase_handle = MockPhaseTHandle(var, obj)
        else:
            phase_handle = MockPhaseTPHandle(var, obj)
        handles.append(phase_handle)
    return Model(handles, var)
    

# %% Energy balance

def iter_T_at_HP(T, H, H_model, phase, mol, P, Cn_model, Cn_cache):
    # Used to solve for temperature at given ethalpy 
    counter, Cn = Cn_cache
    if not counter % 5: Cn_cache[1] = Cn = Cn_model(phase, mol, T, P)
    Cn_cache[0] += 1
    return T + (H - H_model(phase, mol, T, P)) / Cn

def xiter_T_at_HP(T, H, H_model, phase_mol, P, Cn_model, Cn_cache):
    # Used to solve for temperature at given ethalpy 
    counter, Cn = Cn_cache
    if not counter % 5: Cn_cache[1] = Cn = Cn_model(phase_mol, T, P)
    Cn_cache[0] += 1
    return T + (H - H_model(phase_mol, T, P)) / Cn

def iter_T_at_SP(T, S, S_model, phase, mol, P, Cn_model, Cn_cache):
    # Used to solve for temperature at given entropy 
    counter, Cn = Cn_cache
    if not counter % 5: Cn_cache[1] = Cn = Cn_model(phase, mol, T, P)
    Cn_cache[0] += 1
    return T * exp((S - S_model(phase, mol, T, P)) / Cn)

def xiter_T_at_SP(T, S, S_model, phase_mol, P, Cn_model, Cn_cache):
    # Used to solve for temperature at given entropy 
    counter, Cn = Cn_cache
    if not counter % 5: Cn_cache[1] = Cn = Cn_model(phase_mol, T, P)
    Cn_cache[0] += 1
    return T * exp((S - S_model(phase_mol, T, P)) / Cn)

# %% Abstract mixture

stream_mixture_property_interface = {
    'H': ('H', True, False),
    'S': ('S', True, False),
    'C': ('Cn', True, False),
    'Hvap': ('epsilon', True, True),
    'h': ('H', False, False),
    's': ('S', False, False),
    'V': ('V', False, False),
    'Cn': ('Cn', False, False),
    'hvap': ('epsilon', False, True),
    'kappa': ('kappa', False, False),
    'mu': ('mu', False, False),
    'sigma': ('sigma', False, True),
    'epsilon': ('epsilon', False, True),
}

class Mixture:
    """
    Abstract class for estimating mixture properties.
    
    Abstract methods
    ----------------
    Cn(phase, mol, T) : float
        Molar isobaric heat capacity [J/mol/K].
    H(phase, mol, T) : float
        Enthalpy [J/mol].
    S(phase, mol, T, P) : float
        Entropy [J/mol].
    mu(phase, mol, T, P) : float
        Dynamic viscosity [Pa*s].
    V(phase, mol, T, P) : float
        Molar volume [m^3/mol].
    kappa(phase, mol, T, P) : float
        Thermal conductivity [W/m/K].
    Hvap(mol, T) : float
        Heat of vaporization [J/mol]
    sigma(mol, T, P) : float
        Surface tension [N/m].
    epsilon(mol, T, P) : float
        Relative permitivity [-]
    
    """
    MWs: NDArray[float] #: Component molecular weights [g/mol].
    maxiter = 20
    T_tol = 1e-6
    __slots__ = ()
    
    def __call__(self, 
            name, stream=None, *,
            mol=None, T=None, P=None,
            phase=None, phases=None,
        ):
        name, flow, nophase = stream_mixture_property_interface[name]
        if mol is None: mol = stream._imol.data
        else: mol = sparse(mol)
        total = mol.sum()
        if total == 0.:
            return 0. if flow else None
        else:
            if T is None: T = stream.T
            if P is None: P = stream.P
            composition = mol / total
            if mol.ndim == 2:
                if nophase:
                    calculate = getattr(self, name)
                    value = calculate(composition.sum(axis=0), T, P)
                else:
                    calculate = getattr(self, 'x' + name)
                    value = calculate(zip(phases or stream._imol._phases, composition), T, P)
            else:
                calculate = getattr(self, name)
                if nophase:
                    value = calculate(composition, T, P)
                else:
                    value = calculate(phase or stream._imol._phase, composition, T, P)
            return value * total if flow else value  
    
    def MW(self, mol):
        """Return molecular weight [g/mol] given molar array [mol]."""
        if mol.__class__ is not SparseVector: mol = SparseVector(mol)
        total_mol = mol.sum()
        return (self.MWs * mol).sum() / total_mol if total_mol else 0.
    
    def rho(self, phase, mol, T, P):
        """Mixture density [kg/m^3]"""
        MW = self.MW(mol)
        return fn.V_to_rho(self.V(phase, mol, T, P), MW) if MW else 0.
    
    def Cp(self, phase, mol, T):
        """Mixture isobaric heat capacity [J/g/K]"""
        MW = self.MW(mol)
        return self.Cn(phase, mol, T) / MW if MW else 0.
    
    def alpha(self, phase, mol, T, P):
        """Mixture thermal diffusivity [m^2/s]."""
        Cp = self.Cp(phase, mol, T)
        return fn.alpha(self.kappa(phase, mol, T, P), 
                        self.rho(phase, mol, T, P), 
                        Cp * 1000.) if Cp else 0.
    
    def nu(self, phase, mol, T, P):
        """Mixture kinematic viscosity [m^2/s]."""
        rho = self.rho(phase, mol, T, P)
        return fn.mu_to_nu(self.mu(phase, mol, T, P), 
                           rho) if rho else 0.
    
    def Pr(self, phase, mol, T, P):
        """Mixture Prandtl number [-]."""
        Cp = self.Cp(phase, mol, T)
        return fn.Pr(Cp * 1000.,
                     self.kappa(phase, mol, T, P), 
                     self.mu(phase, mol, T, P)) if Cp else 0.
    
    def xrho(self, phase_mol, T, P):
        """Multi-phase mixture density [kg/m3]."""
        return sum([self.rho(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xCp(self, phase_mol, T):
        """Multi-phase mixture isobaric heat capacity [J/g/K]."""
        return sum([self.Cp(phase, mol, T) for phase, mol in phase_mol])
    
    def xalpha(self, phase_mol, T, P):
        """Multi-phase mixture thermal diffusivity [m^2/s]."""
        return sum([self.alpha(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xnu(self, phase_mol, T, P):
        """Multi-phase mixture kinematic viscosity [m^2/s]."""
        return sum([self.nu(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xPr(self, phase_mol, T, P):
        """Multi-phase mixture Prandtl number [-]."""
        return sum([self.Pr(phase, mol, T, P) for phase, mol in phase_mol])
    
    def get_property(self, name, units, *args, **kwargs):
        """
        Return property in requested units.

        Parameters
        ----------
        name : str
            Name of stream property.
        units : str
            Units of measure.
        *args, **kwargs :
            Phase, material and thermal condition.

        """
        value = getattr(self, name)(*args, **kwargs)
        units_dct = thermo_units.chemical_units_of_measure
        if name in units_dct:
            original_units = units_dct[name]
        else:
            raise ValueError(f"'{name}' is not thermodynamic property")
        return original_units.convert(value, units)
    
    def H(self, phase, mol, T, P):
        """Return enthalpy [J/mol]."""
        H = self._H(phase, mol, T, P)
        if self.include_excess_energies: H += self._H_excess(phase, mol, T, P)
        return H
    
    def S(self, phase, mol, T, P):
        """Return entropy in [J/mol/K]."""
        S = self._S(phase, mol, T, P)
        if self.include_excess_energies: S += self._S_excess(phase, mol, T, P)
        return S
    
    def solve_T_at_HP(self, phase, mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        args = (H, self.H, phase, mol, P, self.Cn, [0, None])
        T = iter_T_at_HP(T_guess, *args)
        if abs(T - T_guess) < self.T_tol: return T
        T_guess = flx.aitken(iter_T_at_HP, T, self.T_tol, args, self.maxiter, checkiter=False)
        T = iter_T_at_HP(T_guess, *args)
        return (
            flx.secant(
                lambda T: self.H(phase, mol, T, P) - H,
                x0=T_guess, x1=T, xtol=self.T_tol, ytol=0.
            )
            if abs(T - T_guess) > self.T_tol else T
        )
        
    def xsolve_T_at_HP(self, phase_mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        phase_mol = tuple(phase_mol)
        args = (H, self.xH, phase_mol, P, self.xCn, [0, None])
        T = xiter_T_at_HP(T_guess, *args)
        if abs(T - T_guess) < self.T_tol: return T
        T_guess = flx.aitken(xiter_T_at_HP, T, self.T_tol, args, self.maxiter, checkiter=False)
        T = xiter_T_at_HP(T_guess, *args)
        return (
            flx.secant(
                lambda T: self.xH(phase_mol, T, P) - H,
                x0=T_guess, x1=T, xtol=self.T_tol, ytol=0., checkiter=False
            )
            if abs(T - T_guess) > self.T_tol else T
        )
    
    def solve_T_at_SP(self, phase, mol, S, T_guess, P):
        """Solve for temperature in Kelvin."""
        args = (S, self.S, phase, mol, P, self.Cn, [0, None])
        T = iter_T_at_SP(T_guess, *args)
        if abs(T - T_guess) < self.T_tol: return T
        T_guess = flx.aitken(iter_T_at_SP, T, self.T_tol, args, self.maxiter, checkiter=False)
        T = iter_T_at_SP(T_guess, *args)
        return (
            flx.secant(
                lambda T: self.S(phase, mol, T, P) - S,
                x0=T_guess, x1=T, xtol=self.T_tol, ytol=0.
            )
            if abs(T - T_guess) > self.T_tol else T
        )
        
    def xsolve_T_at_SP(self, phase_mol, S, T_guess, P):
        """Solve for temperature in Kelvin."""
        phase_mol = tuple(phase_mol)
        args = (S, self.xS, phase_mol, P, self.xCn, [0, None])
        T = xiter_T_at_SP(T_guess, *args)
        if abs(T - T_guess) < self.T_tol: return T
        T_guess = flx.aitken(xiter_T_at_SP, T, self.T_tol, args, self.maxiter, checkiter=False)
        T = xiter_T_at_SP(T_guess, *args)
        return (
            flx.secant(
                lambda T: self.xS(phase_mol, T, P) - S,
                x0=T_guess, x1=T, xtol=self.T_tol, ytol=0.
            )
            if abs(T - T_guess) > self.T_tol else T
        )
    
    def xCn(self, phase_mol, T, P=None):
        """Multi-phase mixture molar isobaric heat capacity [J/mol/K]."""
        return sum([self.Cn(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xH(self, phase_mol, T, P):
        """Multi-phase mixture enthalpy [J/mol]."""
        H = self.H
        H_total = sum([H(phase, mol, T, P) for phase, mol in phase_mol])
        return H_total
    
    def xS(self, phase_mol, T, P):
        """Multi-phase mixture entropy [J/mol/K]."""
        S = self.S
        S_total = sum([S(phase, mol, T, P) for phase, mol in phase_mol])
        return S_total
    
    def xV(self, phase_mol, T, P):
        """Multi-phase mixture molar volume [mol/m^3]."""
        return sum([self.V(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xmu(self, phase_mol, T, P):
        """Multi-phase mixture hydrolic [Pa*s]."""
        return sum([self.mu(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xkappa(self, phase_mol, T, P):
        """Multi-phase mixture thermal conductivity [W/m/K]."""
        return sum([self.kappa(phase, mol, T, P) for phase, mol in phase_mol])
        
    def __repr__(self):
        return f"{type(self).__name__}(...)"
    
    def _info(self):
        return (f"{type(self).__name__}(...)")
    
    def show(self):
        print(self._info())
    _ipython_display_ = show


# %% Ideal mixture

class IdealMixture(Mixture):
    """
    Create an Mixture object for estimating mixture properties.
    
    Parameters
    ----------
    Cn : function(phase, mol, T)
        Molar isobaric heat capacity mixture model [J/mol/K].
    H : function(phase, mol, T)
        Enthalpy mixture model [J/mol].
    S : function(phase, mol, T, P)
        Entropy mixture model [J/mol].
    H_excess : function(phase, mol, T, P)
        Excess enthalpy mixture model [J/mol].
    S_excess : function(phase, mol, T, P)
        Excess entropy mixture model [J/mol].
    mu : function(phase, mol, T, P)
        Dynamic viscosity mixture model [Pa*s].
    V : function(phase, mol, T, P)
        Molar volume mixture model [m^3/mol].
    kappa : function(phase, mol, T, P)
        Thermal conductivity mixture model [W/m/K].
    Hvap : function(mol, T)
        Heat of vaporization mixture model [J/mol]
    sigma : function(mol, T, P)
        Surface tension mixture model [N/m].
    epsilon : function(mol, T, P)
        Relative permitivity mixture model [-]
    MWs : 1d array[float]
        Component molecular weights [g/mol].
    include_excess_energies=False : bool
        Whether to include excess energies
        in enthalpy and entropy calculations.
    
    Notes
    -----
    Although the mixture models are on a molar basis, this is only if the molar
    data is normalized before the calculation (i.e. the `mol` parameter is 
    normalized before being passed to the model).
    
    See also
    --------
    IdealTMixtureModel
    IdealTPMixtureModel
    
    Attributes
    ----------
    include_excess_energies : bool
        Whether to include excess energies
        in enthalpy and entropy calculations.
    Cn(phase, mol, T) : 
        Mixture molar isobaric heat capacity [J/mol/K].
    mu(phase, mol, T, P) : 
        Mixture dynamic viscosity [Pa*s].
    V(phase, mol, T, P) : 
        Mixture molar volume [m^3/mol].
    kappa(phase, mol, T, P) : 
        Mixture thermal conductivity [W/m/K].
    Hvap(mol, T, P) : 
        Mixture heat of vaporization [J/mol]
    sigma(mol, T, P) : 
        Mixture surface tension [N/m].
    epsilon(mol, T, P) : 
        Mixture relative permitivity [-].
    MWs : 1d-array[float]
        Component molecular weights [g/mol].
    
    """
    __slots__ = (
        'include_excess_energies',
        'Cn', 'mu', 'V', 'kappa',
        'Hvap', 'sigma', 'epsilon',
        'MWs', '_H', '_H_excess', '_S', '_S_excess',
        '_cache',
    )
    
    def __init__(self, Cn, H, S, H_excess, S_excess,
                 mu, V, kappa, Hvap, sigma, epsilon,
                 MWs, include_excess_energies=False):
        self.include_excess_energies = include_excess_energies
        self.Cn = Cn
        self.mu = mu
        self.V = V
        self.kappa = kappa
        self.Hvap = Hvap
        self.sigma = sigma
        self.epsilon = epsilon
        self.MWs = MWs
        self._H = H
        self._S = S
        self._H_excess = H_excess
        self._S_excess = S_excess
    
    @classmethod
    def from_chemicals(cls, chemicals, 
                       include_excess_energies=False, 
                       cache=True):
        """
        Create a Mixture object from chemical objects.
        
        Parameters
        ----------
        chemicals : Iterable[Chemical]
            For retrieving pure component chemical data.
        include_excess_energies=False : bool
            Whether to include excess energies in enthalpy and entropy calculations.
        rule : str, optional
            Mixing rule. Defaults to 'ideal'.
        cache : optional
            Whether or not to use cached chemicals and cache new chemicals. Defaults to True.
    
        See also
        --------
        :class:`~.mixture.Mixture`
        :class:`~.IdealMixtureModel`
    
        Examples
        --------
        Calculate enthalpy of evaporation for a water and ethanol mixture:
        
        >>> from thermosteam import IdealMixture
        >>> mixture = IdealMixture.from_chemicals(['Water', 'Ethanol'])
        >>> mixture.Hvap([0.2, 0.8], 350)
        39750.62

        Calculate density for a water and ethanol mixture in g/L:

        >>> from thermosteam import Mixture
        >>> mixture = IdealMixture.from_chemicals(['Water', 'Ethanol'])
        >>> mixture.get_property('rho', 'g/L', 'l', [0.2, 0.8], 350, 101325)
        754.23
        
        """
        isa = isinstance
        if isa(chemicals, CompiledChemicals):
            MWs = chemicals.MW
            chemicals = chemicals.tuple
        else:
            chemicals = [(i if isa(i, Chemical) else Chemical(i, cache=cache)) for i in chemicals]
            MWs = chemical_data_array(chemicals, 'MW')
        getfield = getattr
        Cn =  create_mixture_model(chemicals, 'Cn', IdealTMixtureModel)
        H =  create_mixture_model(chemicals, 'H', IdealTPMixtureModel)
        S = create_mixture_model(chemicals, 'S', IdealEntropyModel)
        H_excess = create_mixture_model(chemicals, 'H_excess', IdealTPMixtureModel)
        S_excess = create_mixture_model(chemicals, 'S_excess', IdealTPMixtureModel)
        mu = create_mixture_model(chemicals, 'mu', IdealTPMixtureModel)
        V = create_mixture_model(chemicals, 'V', IdealTPMixtureModel)
        kappa = create_mixture_model(chemicals, 'kappa', IdealTPMixtureModel)
        Hvap = IdealHvapModel(chemicals)
        sigma = SinglePhaseIdealTMixtureModel([getfield(i, 'sigma') for i in chemicals], 'sigma')
        epsilon = SinglePhaseIdealTMixtureModel([getfield(i, 'epsilon') for i in chemicals], 'epsilon')
        return cls(Cn, H, S, H_excess, S_excess,
                   mu, V, kappa, Hvap, sigma, epsilon, MWs, include_excess_energies)
    
    def __repr__(self):
        return f"{type(self).__name__}(..., include_excess_energies={self.include_excess_energies})"
    
    def _info(self):
        return (f"{type(self).__name__}(...\n"
                f"    include_excess_energies={self.include_excess_energies}\n"
                 ")")
    
    
# %% Thermo mixture

class EOSMixture(Mixture):
    __slots__ = (
        'chemicals', 'eos_chemicals', 'Cn_ideal', 'mu', 'V', 
        'kappa', 'sigma', 'epsilon', 'H_ideal', 'S_ideal', 'MWs', 
        'eos_cache', 'active_eos'
    )
    chemsep_db = None
    
    def __init__(self, chemicals, eos_chemicals, 
                 Cn_ideal, H_ideal, S_ideal, 
                 mu, V, kappa, sigma, epsilon, MWs):
        self.chemicals = chemicals
        self.eos_chemicals = eos_chemicals
        self.Cn_ideal = Cn_ideal
        self.H_ideal = H_ideal
        self.S_ideal = S_ideal
        self.mu = mu
        self.V = V
        self.kappa = kappa
        self.sigma = sigma
        self.epsilon = epsilon
        self.MWs = MWs
        self.eos_cache = {}
        self.active_eos = {}
    
    def solve_T_at_HP(self, phase, mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        self.active_eos[phase] = self.eos_args(phase, mol, T_guess, P)
        T = super().solve_T_at_HP(phase, mol, H, T_guess, P)
        self.active_eos.clear()
        return T
        
    def xsolve_T_at_HP(self, phase_mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        phase_mol = tuple(phase_mol)
        for phase, mol in phase_mol: self.active_eos[phase] = self.eos_args(phase, mol, T_guess, P)
        T = super().xsolve_T_at_HP(phase_mol, H, T_guess, P)
        self.active_eos.clear()
        return T
    
    def solve_T_at_SP(self, phase, mol, S, T_guess, P):
        """Solve for temperature in Kelvin."""
        self.active_eos[phase] = self.eos_args(phase, mol, T_guess, P)
        T = super().solve_T_at_SP(phase, mol, S, T_guess, P)
        self.active_eos.clear()
        return T
        
    def xsolve_T_at_SP(self, phase_mol, S, T_guess, P):
        """Solve for temperature in Kelvin."""
        phase_mol = tuple(phase_mol)
        for phase, mol in phase_mol: self.active_eos[phase] = self.eos_args(phase, mol, T_guess, P)
        T = super().xsolve_T_at_SP(phase_mol, S, T_guess, P)
        self.active_eos.clear()
        return T
    
    def eos_args(self, phase, mol, T, P):
        chemicals = self.chemicals
        if mol.__class__ is SparseVector: 
            items = mol.dct.items()
        else:
            items = [(i, j) for i, j in enumerate(mol) if j]
        eos_chemicals = self.eos_chemicals
        chemical_subset = []
        mol_subset = []
        eos_mol = 0
        for i, j in items:
            chemical = chemicals[i]
            if chemical in eos_chemicals:
                chemical_subset.append(chemical)
                mol_subset.append(j)
                eos_mol += j
        zs = [i / eos_mol for i in mol_subset]
        eos_chemicals = tuple(chemical_subset)
        key = eos_chemicals
        cache = self.eos_cache
        only_g = phase == 'g'
        only_l = phase == 'l'
        if key in cache:
            eos = cache[key].to_TP_zs(
                T=T, P=P, zs=zs, only_g=only_g, only_l=only_l,
                fugacities=False
            )
        else:
            data = tmo.ChemicalData(eos_chemicals)
            if self.chemsep_db is None:
                kijs = None
            else:
                try:
                    kijs = IPDB.get_ip_asymmetric_matrix(self.chemsep_db, data.CASs, 'kij')
                except:
                    kijs = None
            self.cache[key] = eos = self.EOS(
                Tcs=data.Tcs, Pcs=data.Pcs, omegas=data.omegas, kijs=kijs,
                T=T, P=P, zs=zs, only_g=only_g, only_l=only_l,
                fugacities=False
            )
        return eos, eos_mol

    def Hvap(self, mol, T, P):
        phase = 'l'
        eos, eos_mol = self.eos_args(phase, mol, T, P)
        return (eos.Hvap(T) * eos_mol).sum()

    def Cn(self, phase, mol, T, P):
        Cn = self.Cn_ideal(phase, mol, T, P)
        if phase != 's':
            if phase in self.active_eos:
                eos, eos_mol = self.active_eos[phase]
            else:
                eos, eos_mol = self.eos_args(
                    phase, mol, T, P
                )
            if phase == 'l':
                try: Cn += eos.Cn_dep_l * eos_mol
                except: 
                    try: Cn += eos.Cn_dep_g * eos_mol
                    except: pass
            else:
                try: Cn += eos.Cn_dep_g * eos_mol
                except: pass
        return Cn
    
    def H(self, phase, mol, T, P):
        """Return enthalpy [J/mol]."""
        H = self.H_ideal(phase, mol, T, P)
        if phase != 's':
            if phase in self.active_eos:
                eos, eos_mol = self.active_eos[phase]
            else:
                eos, eos_mol = self.eos_args(
                    phase, mol, T, P
                )
            if phase == 'l':
                try: H += eos.H_dep_l * eos_mol
                except: 
                    try: H += eos.H_dep_g * eos_mol
                    except: pass
            else:
                try: H += eos.H_dep_g * eos_mol
                except: pass
        return H
    
    def S(self, phase, mol, T, P):
        """Return entropy [J/mol/K]."""
        S = self.S_ideal(phase, mol, T, P)
        if phase != 's':
            if phase in self.active_eos:
                eos, eos_mol = self.active_eos[phase]
            else:
                eos, eos_mol = self.eos_args(
                    phase, mol, T, P
                )
            if phase == 'l':
                try: S += eos.S_dep_l * eos_mol
                except: 
                    try: S += eos.S_dep_g * eos_mol
                    except: pass
            else:
                try: S += eos.S_dep_g * eos_mol
                except: pass
        return S
    
    @classmethod
    def from_chemicals(cls, chemicals, eos_chemicals=None, cache=True):
        """
        Create a EOSMixture object from chemical objects.
        
        Parameters
        ----------
        chemicals : Iterable[Chemical]
            For retrieving pure component chemical data.
        eos_chemicals : Iterable[Chemical]
            Chemicals with equations of state.
        cache : optional
            Whether or not to use cached chemicals and cache new chemicals. Defaults to True.
    
        Examples
        --------
        Calculate enthalpy of evaporation for a water and ethanol mixture:
        
        >>> from thermosteam import PRMixture
        >>> mixture = PRMixture.from_chemicals(['Water', 'Ethanol'])
        >>> mixture.Hvap([0.2, 0.8], 350)
        39750.62

        Calculate density for a water and ethanol mixture in g/L:

        >>> from thermosteam import PRMixture
        >>> mixture = PRMixture.from_chemicals(['Water', 'Ethanol'])
        >>> mixture.get_property('rho', 'g/L', 'l', [0.2, 0.8], 350, 101325)
        754.23
        
        """
        isa = isinstance
        if isa(chemicals, CompiledChemicals):
            MWs = chemicals.MW
            chemicals = chemicals.tuple
        else:
            chemicals = [(i if isa(i, Chemical) else Chemical(i, cache=cache)) for i in chemicals]
            MWs = chemical_data_array(chemicals, 'MW')
        if eos_chemicals is None:
            eos_chemicals = tuple(chemicals)
        getfield = getattr
        Cn = create_mixture_model(chemicals, 'Cn', IdealTMixtureModel)
        H = create_mixture_model(chemicals, 'H', IdealTPMixtureModel)
        S = create_mixture_model(chemicals, 'S', IdealEntropyModel)
        mu = create_mixture_model(chemicals, 'mu', IdealTPMixtureModel)
        V = create_mixture_model(chemicals, 'V', IdealTPMixtureModel)
        kappa = create_mixture_model(chemicals, 'kappa', IdealTPMixtureModel)
        sigma = SinglePhaseIdealTMixtureModel([getfield(i, 'sigma') for i in chemicals], 'sigma')
        epsilon = SinglePhaseIdealTMixtureModel([getfield(i, 'epsilon') for i in chemicals], 'epsilon')
        return cls(chemicals, eos_chemicals, Cn, H, S, mu, V, kappa, sigma, epsilon, MWs)
    
    @classmethod
    def subclass(cls, EOS, name=None):
        if name is None: name = EOS.__name__.replace('MIX', '') + 'Mixture'
        return type(name, (cls,), dict(EOS=EOS, cache={}))

    def _info(self):
        return (
            f"{type(self).__name__}(\n"
            f"    eos_chemicals=[{', '.join([i.ID for i in self.eos_chemicals])}]\n"
             ")"
        )

dct = globals()    
clsnames = []
for name in ('PRMIX', 'SRKMIX', 'PR78MIX', 'VDWMIX', 'PRSVMIX',
             'PRSV2MIX', 'TWUPRMIX', 'TWUSRKMIX', 'APISRKMIX', 'IGMIX', 'RKMIX',
             'PRMIXTranslatedConsistent', 'PRMIXTranslatedPPJP', 'PRMIXTranslated',
             'SRKMIXTranslatedConsistent', 'PSRK', 'MSRKMIXTranslated',
             'SRKMIXTranslated'):
    cls = EOSMixture.subclass(getattr(eos_mix, name))
    clsname = cls.__name__
    clsnames.append(clsname)
    dct[clsname] = cls

dct['PRMixture'].chemsep_db = 'ChemSep PR'
__all__ = (*__all__, *clsnames)
del dct, clsnames


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2025, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
from ..exceptions import NoEquilibrium
from .vle import VLE
from .lle import LLE
from .equilibrium import Equilibrium
from ..utils import Cache
import thermosteam as tmo
import numpy as np
from scipy.optimize import root
from numba import njit

__all__ = ('VLLE', 'VLLECache')

@njit(cache=True)
def RashfordRice_VLLE_residuals(VL1, z, Ka, Kb):
    V, L1 = VL1
    L2 = 1.0 - V - L1
    denom = V + L1 / Ka + L2 / Kb
    y = z / denom
    x_alpha = y / Ka
    residuals = np.zeros(2)
    residuals[0] = y.sum() - 1.0
    residuals[1] = x_alpha.sum() - 1.0
    return residuals

class VLLE(Equilibrium, phases='Llg'):
    """
    Create a VLLE object that performs vapor-liquid-liquid equilibrium when called.
        
    Parameters
    ----------
    imol=None : :class:`~thermosteam.indexer.MaterialIndexer`, optional
        Molar chemical phase data is stored here.
    thermal_condition=None : :class:`~thermosteam.ThermalCondition`, optional
        Temperature and pressure results are stored here.
    thermo=None : :class:`~thermosteam.Thermo`, optional
        Themodynamic property package for equilibrium calculations.
        Defaults to `thermosteam.settings.get_thermo()`.
    
    """
    __slots__ = (
        'top_chemical',
        'vle_l',
        'vle_L',
        'lle',
        'iter',
    )
    
    def __init__(self, 
            imol=None,
            thermal_condition=None,
            thermo=None,
        ):
        super().__init__(imol, thermal_condition, thermo)
        imol = self._imol
        thermo = self._thermo
        thermal_condition = self._thermal_condition
        imol_vle_l = tmo.indexer.MaterialIndexer.from_data(
            imol['gl'], phases=('g', 'l')
        )
        imol_vle_L = tmo.indexer.MaterialIndexer.from_data(
            imol['gL'], phases=('g', 'l')
        )
        imol_lle = tmo.indexer.MaterialIndexer.from_data(
            imol['Ll'], phases=('L', 'l')
        )
        self.vle_l = VLE(imol_vle_l, thermal_condition, thermo)
        self.vle_L = VLE(imol_vle_L, thermal_condition, thermo)
        self.lle = LLE(imol_lle, thermal_condition, thermo)
    
    def __call__(self, *, T=None, P=None, V=None, H=None, S=None, top_chemical=None):
        """
        Perform vapor-liquid-liquid equilibrium.

        Parameters
        ----------
        T : float, optional
            Operating temperature [K].
        P : float, optional
            Operating pressure [Pa].
        V : float, optional
            Molar vapor fraction.
        H : float, optional
            Enthalpy [kJ/hr].
        S : float, optional
            Entropy [kJ/hr/K]
        
        Notes
        -----
        You may only specify two of the following parameters: P, H, T, V, S.
        
        """
        ### Decide what kind of equilibrium to run ###
        T_spec = T is not None
        P_spec = P is not None
        V_spec = V is not None
        H_spec = H is not None
        S_spec = S is not None
        N_specs = (T_spec + P_spec + V_spec + H_spec + S_spec)
        assert N_specs == 2, ("must pass two and only two of the following "
                              "specifications: T, P, V, H, S")
        self.top_chemical = self.chemicals.IDs[self.imol.data.sum(axis=0).argmax()] if top_chemical is None else top_chemical
        # Run equilibrium
        if T_spec:
            if P_spec:
                try:
                    self.set_TP(T, P)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.T = T
                    thermal_condition.P = P
            elif V_spec:
                try:
                    self.set_TV(T, V)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.T = T
            elif H_spec:
                self.set_TH(T, H)
            elif S_spec:
                self.set_TS(T, S)
        elif P_spec:
            if V_spec:
                try:
                    self.set_PV(P, V)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.P = P
            elif H_spec:
                try:
                    self.set_PH(P, H)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.P = P
            elif S_spec:
                try:
                    self.set_PS(P, S)
                except:
                    try:
                        self.set_PS(P, S)
                    except NoEquilibrium:
                        thermal_condition = self._thermal_condition
                        thermal_condition.P = P
        elif S_spec: # pragma: no cover
            if H_spec:
                raise NotImplementedError('specification H and S is invalid')
            else: # V_spec
                raise NotImplementedError('specification V and S not implemented')
        elif H_spec: # pragma: no cover
            raise NotImplementedError('specification V and H not implemented')
    
    def set_TP(self, T, P):
        thermal_condition = self._thermal_condition
        thermal_condition.T = T
        thermal_condition.P = P
        chemicals = self.chemicals
        imol = self.imol
        data = imol['Lgl']
        indices = chemicals.get_vlle_indices(data.nonzero_keys())
        subdata = data[:, indices]
        total = subdata.sum()
        subdata = subdata / total
        z = subdata.sum(axis=0)
        self.iter = 0
        new_subdata = flx.fixed_point(
            self._iter_flows_at_TP, subdata, xtol=1e-6, 
            args=(T, P, indices, z),
            convergenceiter=10, 
            checkconvergence=False,
            checkiter=False,
            maxiter=100,
        )
        data[:, indices] = total * new_subdata * z / new_subdata.sum(axis=0)
    
    def _iter_flows_at_TP(self, data, T, P, indices, z):
        imol = self.imol
        L_mol, g_mol, l_mol = imol['Lgl']
        data *= z / data.sum(axis=0)
        L_mol[indices] = data[0]
        g_mol[indices] = data[1]
        l_mol[indices] = data[2]
        self.iter += 1
        
        # LLE
        try:
            self.lle(T=T, P=P, top_chemical=self.top_chemical, use_cache=True)
        except:
            self.lle(T=T, P=P, top_chemical=self.top_chemical)
        
        # VLE with extract
        try:
            self.vle_L._setup()
            self.vle_L._solve_TP(T=T, P=P)
        except:
            self.vle_L(T=T, P=P)
        y = g_mol[indices]
        V0 = y.sum()
        xa = L_mol[indices]
        La = xa.sum()
        if La and V0: 
            y[y < 1e-16] = 1e-16
            xa[xa < 1e-16] = 1e-16
            y /= V0
            xa /= La
            Ka = y / xa
        else:
            Ka = None
        
        # VLE with raffinate
        try:
            self.vle_l._setup()
            self.vle_l._solve_TP(T=T, P=P)
        except:
            self.vle_l(T=T, P=P)
        y = g_mol[indices]
        V1 = y.sum() 
        xb = l_mol[indices]
        Lb = xb.sum()
        if V1 and Lb:
            y[y < 1e-16] = 1e-16
            xb[xb < 1e-16] = 1e-16
            y /= V1
            xb /= Lb
            Kb = y / xb
        else:
            Kb = None
        
        if Ka is None or Kb is None:
            # At most 2 phases
            values = np.array([
                L_mol[indices],
                g_mol[indices],
                l_mol[indices],
            ])
        else:
            # Potentially 3 phases
            guess = np.array([V1, La, Lb])
            guess /= guess.sum()
            sol = root(
                RashfordRice_VLLE_residuals, 
                guess[:2], 
                args=(z, Ka, Kb), 
                method="hybr", 
                tol=1e-12,
            )
            V, La = sol.x
            Lb = 1 - V - La
            if V < 1e-12 or La < 1e-12 or Lb < 1e-12:
                # Actually 2 phases
                values = np.array([
                    L_mol[indices],
                    g_mol[indices],
                    l_mol[indices],
                ])
            else:
                denom = V + La / Ka + Lb / Kb
                y = z / denom
                y /= y.sum()
                xa = y / Ka
                xa /= xa.sum()
                xb  = y / Kb
                xb /= xb.sum()
                values = np.array([
                    xa * La, y * V, xb * Lb
                ])
        values[values < 1e-12] = 0
        return values
    
    # def set_TP(self, T, P):
    #     thermal_condition = self._thermal_condition
    #     thermal_condition.T = T
    #     thermal_condition.P = P
    #     chemicals = self.chemicals
    #     imol = self.imol
    #     data = imol['Lgl']
    #     indices = chemicals.get_vlle_indices(data.nonzero_keys())
    #     subdata = data[:, indices]
    #     total = subdata.sum()
    #     subdata = subdata / total
    #     z = subdata.sum(axis=0)
        
    #     Ka, Kb = self._evaluate_Ks(z, T, P, indices, z)
    #     if Ka is None or Kb is None:
    #         new_values = data[:, indices]
    #         data[:, indices] *= total * new_values * z / new_values.sum(axis=0)
    #     else:
    #         self.iter = 0
    #         logKs = np.log([Ka, Kb])
    #         new_subdata = flx.fixed_point(
    #             self._iter_flows_at_TP, logKs, xtol=1e-6, 
    #             args=(T, P, indices, z),
    #             convergenceiter=10, 
    #             checkconvergence=False,
    #             checkiter=False,
    #             maxiter=100,
    #         )
    #         data[:, indices] = total * new_subdata * z / new_subdata.sum(axis=0)
    
    # def _evaluate_Ks(self, data, T, P, indices, z):
    #     imol = self.imol
    #     L_mol, g_mol, l_mol = imol['Lgl']
    #     data *= z / data.sum(axis=0)
    #     L_mol[indices] = data[0]
    #     g_mol[indices] = data[1]
    #     l_mol[indices] = data[2]
    #     self.iter += 1
        
    #     # LLE
    #     try:
    #         self.lle(T=T, P=P, top_chemical=self.top_chemical, use_cache=True)
    #     except:
    #         self.lle(T=T, P=P, top_chemical=self.top_chemical)
        
    #     # VLE with extract
    #     try:
    #         self.vle_L._setup()
    #         self.vle_L._solve_TP(T=T, P=P)
    #     except:
    #         self.vle_L(T=T, P=P)
    #     y = g_mol[indices]
    #     V0 = y.sum()
    #     xa = L_mol[indices]
    #     La = xa.sum()
    #     if La and V0: 
    #         y[y < 1e-16] = 1e-16
    #         xa[xa < 1e-16] = 1e-16
    #         y /= V0
    #         xa /= La
    #         Ka = y / xa
    #     else:
    #         Ka = None
        
    #     # VLE with raffinate
    #     try:
    #         self.vle_l._setup()
    #         self.vle_l._solve_TP(T=T, P=P)
    #     except:
    #         self.vle_l(T=T, P=P)
    #     y = g_mol[indices]
    #     V1 = y.sum() 
    #     xb = l_mol[indices]
    #     Lb = xb.sum()
    #     if V1 and Lb:
    #         y[y < 1e-16] = 1e-16
    #         xb[xb < 1e-16] = 1e-16
    #         y /= V1
    #         xb /= Lb
    #         Kb = y / xb
    #     else:
    #         Kb = None
        
    #     return Ka, Kb
    
    def _T_bubble_iter(self, T):
        thermal_condition = self._thermal_condition
        self.lle(
            T=T,
            P=thermal_condition.P,
            top_chemical=self.top_chemical
        )
        self.vle_L(V=0, P=thermal_condition.P)
        return thermal_condition.T
     
    def _P_bubble_iter(self, P):
        thermal_condition = self._thermal_condition
        self.lle(
            T=thermal_condition.T,
            P=P,
            top_chemical=self.top_chemical
        )
        self.vle_L(V=0, P=P)
        return thermal_condition.P
       
    def bubble_point_at_T(self, T=None):
        thermal_condition = self._thermal_condition
        if T is None: 
            T = thermal_condition.T
        else:
            thermal_condition.T = T
        imol = self.imol
        l_mol = imol['l']
        g_mol = imol['g']
        l_mol += g_mol
        g_mol[:] = 0
        return flx.wegstein(
            self._P_bubble_iter, 
            thermal_condition.P,
            xtol=1e-6, 
            convergenceiter=10, 
            checkconvergence=False,
            checkiter=False,
            maxiter=100,
        )
    
    def dew_point_at_T(self, T=None):
        thermal_condition = self.thermal_condition
        imol = self.imol
        imol['g'] += imol['L'] + imol['l']
        imol['l'].clear()
        imol['L'].clear()
        if T is None: 
            T = thermal_condition.T
        else:
            thermal_condition.T = T
        self.vle_l(V=1, T=T)
        return thermal_condition.P
    
    def bubble_point_at_P(self, P=None):
        thermal_condition = self._thermal_condition
        if P is None: 
            P = thermal_condition.P
        else:
            thermal_condition.P = P
        imol = self.imol
        l_mol = imol['l']
        g_mol = imol['g']
        l_mol += g_mol
        g_mol[:] = 0
        return flx.wegstein(
            self._T_bubble_iter, 
            thermal_condition.T,
            xtol=1e-6, 
            convergenceiter=10, 
            checkconvergence=False,
            checkiter=False,
            maxiter=100,
        )
    
    def dew_point_at_P(self, P=None):
        thermal_condition = self.thermal_condition
        imol = self.imol
        imol['g'] += imol['L'] + imol['l']
        imol['l'].clear()
        imol['L'].clear()
        if P is None: 
            P = thermal_condition.P
        else:
            thermal_condition.P = P
        self.vle_l(V=1, P=P)
        return thermal_condition.T
    
    def _H_hat_err_at_T(self, T, H_hat, F_mass, phase_data):
        P = self.thermal_condition.P
        self.set_TP(T, P)
        return self.mixture.xH(phase_data, T, P) / F_mass - H_hat
    
    def _H_hat_err_at_P(self, P, H_hat, F_mass, phase_data):
        T = self.thermal_condition.T
        self.set_TP(T, P)
        return self.mixture.xH(phase_data, T, P) / F_mass - H_hat
    
    @property
    def vapor_fraction(self):
        imol = self.imol
        g = imol['g'].sum()
        return g / (g + imol['lL'].sum())
    
    def _V_hat_err_at_T(self, T, V):
        P = self.thermal_condition.P
        self.set_TP(T, P)
        return self.vapor_fraction - V
    
    def _V_hat_err_at_P(self, P, V):
        T = self.thermal_condition.T
        self.set_TP(T, P)
        return self.vapor_fraction - V
    
    def _S_hat_err_at_T(self, T, H_hat, F_mass, phase_data):
        P = self.thermal_condition.P
        self.set_TP(T, P)
        return self.mixture.xS(phase_data, T, P) / F_mass - H_hat
    
    def _S_hat_err_at_P(self, P, H_hat, F_mass, phase_data):
        T = self.thermal_condition.T
        self.set_TP(T, P)
        return self.mixture.xS(phase_data, T, P) / F_mass - H_hat
    
    def set_PH(self, P, H):
        thermal_condition = self.thermal_condition
        thermal_condition.P = P
        T_guess = thermal_condition.T
        imol = self.imol
        phase_data = tuple(imol)
        
        # Check if subcooled liquid
        T_bubble = self.bubble_point_at_P(P)
        H_bubble = self.mixture.xH(phase_data, T_bubble, P)
        dH_bubble = H - H_bubble
        if dH_bubble <= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_HP(phase_data, H, T_bubble, P)
            return
        
        # Check if super heated vapor
        T_dew = self.dew_point_at_P(P)
        if T_dew <= T_bubble: 
            T_dew, T_bubble = T_bubble, T_dew
            T_dew += 0.5
            T_bubble -= 0.5
        H_dew = self.mixture.xH(phase_data, T_dew, P)
        dH_dew = H - H_dew
        if dH_dew >= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_HP(phase_data, H, T_dew, P)
            return
        
        F_mass = self.mixture.MW(imol.data.sum(axis=0))
        H_hat = H / F_mass
        H_hat_bubble = H_bubble / F_mass
        H_hat_dew = H_dew / F_mass
        T0 = flx.IQ_interpolation(
            self._H_hat_err_at_T, T_bubble, T_dew, 
            H_hat_bubble - H_hat, H_hat_dew - H_hat,
            T_guess, 1e-3, 1e-3,
            (H_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=10,
        )
        T1 = self.mixture.xsolve_T_at_HP(phase_data, H, thermal_condition.T, P)
        T = flx.IQ_interpolation(
            self._H_hat_err_at_T, 2 * T0 - T1, T1, 
            xtol=1e-6, ytol=1e-6,
            args=(H_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=100,
        )
        self.thermal_condition.T = self.mixture.xsolve_T_at_HP(phase_data, H, T, P)

    def set_TH(self, T, H):
        thermal_condition = self.thermal_condition
        thermal_condition.T = T
        P_guess = thermal_condition.P
        imol = self.imol
        phase_data = tuple(imol)
        
        # Check if subcooled liquid
        P_bubble = self.bubble_point_at_T(T)
        H_bubble = self.mixture.xH(phase_data, T, P_bubble)
        dH_bubble = H - H_bubble
        if dH_bubble <= 0:
            raise NotImplementedError('cannot solve for pressure yet')
        
        # Check if super heated vapor
        P_dew = self.dew_point_at_T(T)
        if P_bubble <= P_dew: 
            P_dew, P_bubble = P_bubble, P_dew
            P_dew += 0.5
            P_bubble -= 0.5
        H_dew = self.mixture.xH(phase_data, T, P_dew)
        dH_dew = H - H_dew
        if dH_dew >= 0:
            raise NotImplementedError('cannot solve for pressure yet')
        
        F_mass = self.mixture.MW(imol.data.sum(axis=0))
        H_hat = H / F_mass
        H_hat_bubble = H_bubble / F_mass
        H_hat_dew = H_dew / F_mass
        flx.IQ_interpolation(
            self._H_hat_err_at_P, P_bubble, P_dew, 
            H_hat_bubble - H_hat, H_hat_dew - H_hat,
            P_guess, 1, 1e-6,
            (H_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=100,
        )

    def set_PS(self, P, S):
        thermal_condition = self.thermal_condition
        thermal_condition.P = P
        T_guess = thermal_condition.T
        imol = self.imol
        phase_data = tuple(imol)
        
        # Check if subcooled liquid
        T_bubble = self.bubble_point_at_P(P)
        S_bubble = self.mixture.xS(phase_data, T_bubble, P)
        dS_bubble = S - S_bubble
        if dS_bubble <= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_SP(phase_data, S, T_bubble, P)
            return
        
        # Check if super heated vapor
        T_dew = self.dew_point_at_P(P)
        if T_dew <= T_bubble: 
            T_dew, T_bubble = T_bubble, T_dew
            T_dew += 0.5
            T_bubble -= 0.5
        S_dew = self.mixture.xS(phase_data, T_dew, P)
        dS_dew = S - S_dew
        if dS_dew >= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_SP(phase_data, S, T_dew, P)
            return
        
        F_mass = self.mixture.MW(imol.data.sum(axis=0))
        S_hat = S / F_mass
        S_hat_bubble = S_bubble / F_mass
        S_hat_dew = S_dew / F_mass
        T0 = flx.IQ_interpolation(
            self._S_hat_err_at_T, T_bubble, T_dew, 
            S_hat_bubble - S_hat, S_hat_dew - S_hat,
            T_guess, 1e-3, 1e-3,
            (S_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=10,
        )
        T1 = self.mixture.xsolve_T_at_SP(phase_data, S, thermal_condition.T, P)
        T = flx.IQ_interpolation(
            self._S_hat_err_at_T, 2 * T0 - T1, T1, 
            xtol=1e-6, ytol=1e-6,
            args=(S_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=100,
        )
        self.thermal_condition.T = self.mixture.xsolve_T_at_SP(phase_data, S, T, P)
        
    def set_TS(self, T, S):
        thermal_condition = self.thermal_condition
        thermal_condition.T = T
        P_guess = thermal_condition.P
        imol = self.imol
        phase_data = tuple(imol)
        
        # Check if subcooled liquid
        P_bubble = self.bubble_point_at_T(T)
        S_bubble = self.mixture.xS(phase_data, T, P_bubble)
        dS_bubble = S - S_bubble
        if dS_bubble <= 0:
            thermal_condition.P = self.mixture.xsolve_P_at_ST(phase_data, S, P_bubble, T)
            return
        
        # Check if super heated vapor
        P_dew = self.dew_point_at_T(T)
        if P_bubble <= P_dew: 
            P_dew, P_bubble = P_bubble, P_dew
            P_dew += 0.5
            P_bubble -= 0.5
        S_dew = self.mixture.xS(phase_data, P_dew, T)
        dS_dew = S - S_dew
        if dS_dew >= 0:
            thermal_condition.P = self.mixture.xsolve_P_at_ST(phase_data, S, P_dew, T)
            return
        
        F_mass = self.mixture.MW(imol.data.sum(axis=0))
        S_hat = S / F_mass
        S_hat_bubble = S_bubble / F_mass
        S_hat_dew = S_dew / F_mass
        flx.IQ_interpolation(
            self._S_hat_err_at_P, P_bubble, P_dew, 
            S_hat_bubble - S_hat, S_hat_dew - S_hat,
            P_guess, 1, 1e-6,
            (S_hat, F_mass, phase_data),
            checkiter=False, checkbounds=False,
            maxiter=100,
        )
        
    def set_PV(self, P, V):
        thermal_condition = self.thermal_condition
        thermal_condition.P = P
        if V == 0:
            self.bubble_point_at_P(P)
        elif V == 1:
            self.dew_point_at_P(P)
        else:
            T_bubble = self.bubble_point_at_P(P)
            T_dew = self.dew_point_at_P(P)
            flx.IQ_interpolation(
                self._V_hat_err_at_T, T_bubble, T_dew, 
                thermal_condition.T, xtol=1e-6, ytol=1e-6, args=(V,),
                checkiter=False, checkbounds=False,
                maxiter=100,
            )

    def set_TV(self, T, V):
        thermal_condition = self.thermal_condition
        thermal_condition.T = T
        if V == 0:
            self.bubble_point_at_T(T)
        elif V == 1:
            self.dew_point_at_T(T)
        else:
            P_bubble = self.bubble_point_at_T(T)
            P_dew = self.dew_point_at_T(T)
            flx.IQ_interpolation(
                self._V_hat_err_at_P, P_bubble, P_dew, 
                x=thermal_condition.P, xtol=1, ytol=1e-6, args=(V,),
                checkiter=False, checkbounds=False,
                maxiter=100,
            )
    
class VLLECache(Cache): load = VLLE
del Cache, Equilibrium

